#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    Sample bot that translates text whenever it sees a message with one of its keywords.
"""

import argparse
import logging
import sys
import os
import shutil
import json
import i18n
import re
import locale
import requests
import random
# Provides an easy way to get the unicode sequence for country flags
import flag
import yaml
import urllib.request

# Own classes
from .helpers import *
from .bot import Bot
from .bot import ArgsHelper as BotArgsHelper
from .console import ConsoleChatter
from .jabber import JabberChatter
from .jabber import arg_parser as jabber_arg_parser
from .signalcli import SignalChatter
from .signalcli import ArgsHelper as SignalArgsHelper
from .stealth import StealthChatter



# Default timeout for requests in seconds
# Note : More than 10s recommended (30s ?) on IBM Cloud with a free account
TIMEOUT = 60

# Set to None to translate keywords in all available languages
# Set to something > 0 to limit the number of translations for the keywords (for tests)
LIMIT_KEYWORDS = None

# See https://github.com/nicolabs/nicobot/issues/8
# Description : https://unicode.org/reports/tr35/#Likely_Subtags
# Original XML version : http://cldr.unicode.org/index/cldr-spec/language-tag-equivalences
# This is the URL to the JSON version
LIKELY_SUBTAGS_URL = "https://raw.githubusercontent.com/unicode-cldr/cldr-core/master/supplemental/likelySubtags.json"


# Default configuration (some defaults still need to be set up after command line has been parsed)
class Config:

    def __init__(self):
        self.__dict__.update({
            'backend': "console",
            'config_file': None,
            'config_dir': os.getcwd(),
            'group': None,
            'ibmcloud_url': None,
            'ibmcloud_apikey': None,
            'input_file': sys.stdin,
            'keywords': [],
            'keywords_files': [],
            'languages': [],
            'languages_file': None,
            'languages_likely': None,
            # e.g. locale.getlocale() may return ('en_US','UTF-8') : we only keep the 'en_US' part here (the same as the expected command-line parameter)
            'locale': locale.getlocale()[0],
            'recipient': None,
            'shutdown': None,
            'signal_cli': shutil.which("signal-cli"),
            'signal_stealth': False,
            'stealth': False,
            'username': None,
            'verbosity': "WARNING"
            })


"""
    TODO Find a better way to log requests.Response objects
"""
def _logResponse( r ):
    logging.debug("<<< Response : %s\tbody: %s", repr(r), r.content )


def sanitizeNotPattern( string ):
    """
        Returns a string that will only all 'non-word' characters escaped with backslash
        in order to use it in a regular expression pattern without including special character.

        We could just replace any character 'c' with '\c' but replacing only special characters keep it somewhat still readable.
    """
    return re.sub( r'([^\w])', '\\\\\\1', string )



class TransBot(Bot):
    """
        Sample bot that translates text.

        It only answers to messages containing defined keywords.
        It uses IBM Watson™ Language Translator (see API docs : https://cloud.ibm.com/apidocs/language-translator) to translate the text.
    """


    def __init__( self,
        chatter, ibmcloud_url, ibmcloud_apikey,
        keywords=[], keywords_files=[],
        languages=[], languages_file=None, languages_likely=None,
        locale=re.split(r'[_-]',locale.getlocale()[0]),
        shutdown_pattern=r'bye nicobot' ):
        """
            keywords: list of keywords that will trigger this bot (in any supported language)
            keywords_files: list of JSON files with each a list of keywords (or write into)
            languages: List of supported languages in this format : https://cloud.ibm.com/apidocs/language-translator#list-identifiable-languages
            languages_file: JSON file where to find the list of target languages (or write into)
            languages_likely: JSON URI where to find Unicode's likely subtags (or write into)
            locale: overrides the default locale ; tuple like : ('en','GB')
            shutdown_pattern: a regular expression pattern that terminates this bot
            chatter: the backend chat engine
            ibmcloud_url (required): IBM Cloud API base URL (e.g. 'https://api.eu-de.language-translator.watson.cloud.ibm.com/instances/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxx')
            ibmcloud_apikey (required): IBM Cloud API key (e.g. 'dG90byBlc3QgZGFucyBsYSBwbGFjZQo')
            store_path: Base directory where to cache files
        """

        self.status = {'events':[]}

        self.ibmcloud_url = ibmcloud_url
        self.ibmcloud_apikey = ibmcloud_apikey
        self.chatter = chatter

        self.locale = locale
        self.languages = languages
        if languages_file:
            # Only after IBM credentials have been set can we retrieve the list of supported languages
            self.languages = self.loadLanguages(file=languages_file,locale=locale[0])
        # How many different languages to try to translate to
        self.tries = 5

        self.likelyLanguages = self.loadLikelyLanguages(languages_likely)

        # After self.languages has been set, we can iterate over it to translate keywords
        kws = self.loadKeywords( keywords=keywords, files=keywords_files, limit=LIMIT_KEYWORDS )
        # And build a regular expression pattern with all keywords and their translations
        pattern = r'\b%s\b' % sanitizeNotPattern(kws[0])
        for keyword in kws[1:]:
            pattern = pattern + r'|\b%s\b' % sanitizeNotPattern(keyword)
        # Built regular expression pattern that triggers an answer from this bot
        self.re_keywords = pattern
        # Regular expression pattern of messages that stop the bot
        self.re_shutdown = shutdown_pattern


    def _logEvent( self, event ):

        self.status['events'].append(event)


    def loadLanguages( self, force=False, file=None, locale='en' ):
        """
            Loads the list of known languages.

            Requires the IBM Cloud credentials to be set before !

            If force==True then calls the remote service, otherwise reads from the given file if given
        """

        # TODO It starts with the same code as in loadKeywords : make it a function

        # Gets the list from a local file
        if not force and file:
            logging.debug("Reading from %s..." % file)
            try:
                with open(file,'r') as f:
                    j = json.load(f)
                    return j['languages']
            except:
                logging.info("Could not read languages list from %s" % file)
                pass

        # Else, gets the list from the cloud
        # curl --user apikey:{apikey} "{url}/v3/identifiable_languages?version=2018-05-01"
        url = "%s/v3/identifiable_languages?version=2018-05-01" % self.ibmcloud_url
        headers = {
            'Accept': 'application/json',
            'X-Watson-Learning-Opt-Out': 'true'
            }
        # FIXME Since IBM API doesn't support an Accept-Language header to get the languages name in the locale, we need to query it again
        logging.debug(">>> GET %s, %s",url,repr(headers))
        r = requests.get(url, headers=headers, auth=('apikey',self.ibmcloud_apikey), timeout=TIMEOUT)
        _logResponse(r)

        if r.status_code == requests.codes.ok:
            languages_root = r.json()
            languages = languages_root['languages']

            # IBM Cloud always returns language names in english
            # So we need to translate them if the locale is different
            if locale != 'en':
                languages_names = [ l['name'] for l in languages ]
                translations = self.translate(languages_names,source='en',target=locale)
                logging.debug("Got the following translations for languages names : %s",repr(translations))
                # From my tests seems that IBM cloud returns the original text if it could not translate it
                # so the output list will always be the same size as the input one
                t = 0
                for language in languages:
                    language['name'] = translations['translations'][t]['translation'].strip()
                    t = t + 1

            # Save it for the next time
            if file:
                try:
                    logging.debug("Saving languages to %s..." % file)
                    with open(file,'w') as f:
                        json.dump(languages_root,f)
                except:
                    logging.exception("Could not save the languages list to %s" % file)
                    pass
            else:
                logging.debug("Not saving languages as no file was given")

            return languages
        else:
            r.raise_for_status()


    def loadKeywords( self, keywords=[], files=[], limit=None ):
        """
            Generates a list of translations from a list of keywords.

            Requires self.languages to be filled before !

            If 'keywords' is not empty, will download the translations from IBM Cloud
            and if a single 'file' was given, will save them into it.
            Otherwise, will read from all the given 'files'
        """

        # TODO It starts with the same code as in loadLanguages : make it a function

        kws = []

        # Gets the list from a local file
        if len(keywords) == 0:
            for file in files:
                logging.debug("Reading from %s..." % file)
                # May throw an error
                with open(file,'r') as f:
                    kws = kws + json.load(f)
            logging.debug("Read keyword list : %s",repr(kws))
            return kws

        # TODO remove duplicates
        for keyword in keywords:
            logging.debug("Init %s...",keyword)
            kws = kws + [ keyword ]

            for lang in self.languages:
                # For tests, in order not to use all credits, we can limit the number of calls here
                if limit and len(kws) >= limit:
                    break
                try:
                    translation = self.translate( [keyword], target=lang['language'] )
                    if translation:
                        for t in translation['translations']:
                            translated = t['translation'].strip()
                            logging.debug("Adding translation %s in %s for %s", t, lang, keyword)
                            kws = kws + [ translated ]
                except:
                    logging.exception("Could not translate %s into %s", keyword, repr(lang))
                    pass
        logging.debug("Keywords : %s", repr(kws))

        # TODO ? Save the translations for each keyword into a separate file ?
        if files and len(files) == 1:
            try:
                logging.debug("Saving keywords translations into %s...", files[0])
                with open(files[0],'w') as f:
                    json.dump(kws,f)
            except:
                logging.exception("Could not save keywords translations into %s", files[0])
                pass
        else:
            logging.debug("Not saving keywords as a (single) file was not given")

        return kws


    def loadLikelyLanguages( self, file ):
        """
            Returns a dict from a Likely Subtags JSON structure in the given file.
            If the file cannot be read, will download it from LIKELY_SUBTAGS_URL and save it with the given filename.
        """

        try:
            logging.debug("Loading likely languages from %s",file)
            with open(file,'r') as f:
                return json.load(f)
        except:
            logging.debug("Downloading likely subtags from %s",LIKELY_SUBTAGS_URL)
            with urllib.request.urlopen(LIKELY_SUBTAGS_URL) as response:
                likelySubtags = response.read()
                logging.log(TRACE,"Got likely subtags : %s",repr(likelySubtags))
                # Saves it for the next time
                try:
                    logging.debug("Saving likely subtags into %s",file)
                    with open(file,'w') as f:
                        f.write(likelySubtags.decode())
                except:
                    logging.exception("Error saving the likely languages into %s",repr(file))
                return json.loads(likelySubtags)


    def translate( self, messages, target, source=None ):
        """
            Translates a given list of messages.

            target: Target language short code (e.g. 'en')
            source: Source language short code ; if not given will try to guess

            Returns the full JSON translation as per the IBM cloud service or None if no translation could be found.
        """

        # curl -X POST -u "apikey:{apikey}" --header "Content-Type: application/json" --data "{\"text\": [\"Hello, world! \", \"How are you?\"], \"model_id\":\"en-es\"}" "{url}/v3/translate?version=2018-05-01"
        url = "%s/v3/translate?version=2018-05-01" % self.ibmcloud_url
        body = {
            "text": messages,
            "target": target
            }
        if source:
            body['source'] = source
        headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'X-Watson-Learning-Opt-Out': 'true'
            }
        logging.debug(">>> POST %s, %s, %s",url,repr(body),repr(headers))
        r = requests.post(url, json=body, headers=headers, auth=('apikey',self.ibmcloud_apikey), timeout=TIMEOUT)
        # TODO Log full response when it's usefull (i.e. when a message is going to be answered)
        _logResponse(r)
        if r.status_code == requests.codes.ok:
            return r.json()
        # A 404 can happen if there is no translation available
        elif r.status_code == requests.codes.not_found:
            return None
        else:
            r.raise_for_status()


    def languageToCountry( self, lang ):
        """
            Returns the most likely ISO 3361 country code from an (~ISO 639 or IBM-custom) language
            or the given 'lang' if no country code could be identified.

            lang : the language returned by IBM Translator service (is it ISO 639 ?)

            See https://github.com/nicolabs/nicobot/issues/8
            Likely subtags explanation and format :
            - https://unicode.org/reports/tr35/#Likely_Subtags
            - http://cldr.unicode.org/index/cldr-spec/language-tag-equivalences
        """
        try:
            aa_Bbbb_CC = self.likelyLanguages['supplemental']['likelySubtags'][lang]
            logging.log(TRACE,"Found likely subtags %s for language %s",aa_Bbbb_CC,lang)
            # The last part is the ISO 3361 country code
            return re.split( r'[_-]', aa_Bbbb_CC )[-1]
        except:
            logging.warning("Could not find a country code for %s : returning itself",lang, exc_info=True)
            return lang


    def formatTranslation( self, translation, target ):
        """
            Common decoration of translated messages

            transation = the result of translate()
            target = reminder of which target language was asked (does not appear in the response of translate())
        """

        text = translation['translations'][0]['translation'].strip()
        try:
            # Note : translation['detected_language'] is the detected source language, if guessed
            country = self.languageToCountry(target)
            lang_emoji = flag.flag(country)
        except ValueError:
            logging.debug("Error looking for flag %s",target,exc_info=True)
            lang_emoji= "🏳️‍🌈"
        answer = "%s %s" % (text,lang_emoji)
        return i18n.t('all_messages',message=answer)


    def identifyLanguage( self, language_name ):
        """
            Finds the language code from its name
        """
        # TODO should be at 'trace' level
        logging.debug("identifyLanguage(%s)",language_name)

        # First checks if this is already the language's code (more accurate)
        if language_name in [ l['language'] for l in self.languages ]:
            logging.debug("Identified language is already a code : %s",language_name)
            return language_name
        # Else, really try with the language's name
        else:
            matching_names = [ l for l in self.languages if re.search(language_name.strip(),l['name'],re.IGNORECASE) ]
            logging.debug("Identified languages by name : %s",matching_names)
            if len(matching_names) > 0:
                # Only take the first one
                return matching_names[0]['language']
            else:
                logging.warning("Could not identify language %s",language_name)
                return None


    def onMessage( self, message ):
        """
            Called by self.chatter whenever a message hsa arrived :
            if the given message contains any of the keywords in any language,
            will answer with a translation in a random language
            including the flag of the random language.

            message: A plain text message
            Returns nothing
        """
        logging.debug("onMessage(%s)",message)

        # Preparing the 'translate a message' case
        to_lang = self.locale[0]
        matched_translate = re.search( i18n.t('translate'), message.strip(), flags=re.IGNORECASE )
        # Case where the target language is given
        if matched_translate:
            logging.debug("Detected 'translate a message with target' case")
            to_lang = self.identifyLanguage( matched_translate.group('language') )
            logging.debug("Found target language in message : %s"%to_lang)
        # Case where the target language is not given ; we will simply use the current locale
        else:
            matched_translate = re.search( i18n.t('translate_default_locale'), message.strip(), flags=re.IGNORECASE )
            if matched_translate:
                logging.debug("Detected 'translate a message' case")

        ###
        #
        # Case 'shutdown'
        #
        # FIXME re.compile((i18n.t('Shutdown'),re.IGNORECASE).search(message) does not work
        # as expected so we use re.search(...)
        if re.search( self.re_shutdown, message, re.IGNORECASE ):
            logging.debug("Shutdown asked")
            self._logEvent({ 'type':'shutdown', 'message':message })
            self.chatter.stop()

        ###
        #
        # Case 'translate a message'
        #
        elif matched_translate:
            status_event = { 'type':'translate', 'message':message, 'target_lang':to_lang }
            self._logEvent(status_event)
            if to_lang:
                translation = self.translate( [matched_translate.group('message')],target=to_lang )
                logging.debug("Got translation : %s",repr(translation))
                status_event['translation'] = translation
                if translation and len(translation['translations'])>0:
                    answer = self.formatTranslation(translation,target=to_lang)
                    logging.debug(">> %s" % answer)
                    status_event['answer'] = answer
                    self.chatter.send(answer)
                else:
                    # TODO Make translate throw an error with details
                    logging.warning("Did not get a translation in %s for %s",to_lang,message)
                    answer = i18n.t('all_messages',message=i18n.t('IDontKnow'))
                    status_event['error'] = 'no_translation'
                    status_event['answer'] = answer
                    self.chatter.send(answer)
            else:
                logging.warning("Could not identify target language in %s",message)
                answer = i18n.t('all_messages',message=i18n.t('IDontKnow'))
                status_event['error'] = 'unknown_target_language'
                status_event['answer'] = answer
                self.chatter.send( i18n.t('all_messages',message=i18n.t('IDontKnow')) )

        ###
        #
        # Case 'answer to keywords'
        #
        elif re.search( self.re_keywords, message, flags=re.IGNORECASE ):

            status_translations = []
            status_event = { 'type':'keyword', 'message':message, 'translations':status_translations }
            self._logEvent( status_event )

            # Selects a few random target languages each time
            langs = random.choices( self.languages, k=self.tries )

            for lang in langs:
                # Gets a translation in this random language
                translation = self.translate( [message], target=lang['language'] )
                logging.debug("Got translation : %s",repr(translation))
                status_translation = { 'target_language':lang['language'], 'translation':translation }
                status_translations.append(status_translation)
                if translation and len(translation['translations'])>0:
                    answer = self.formatTranslation(translation,target=lang['language'])
                    logging.debug(">> %s" % answer)
                    status_translation['answer'] = answer
                    self.chatter.send(answer)
                    # Returns as soon as one translation was done
                    return
                else:
                    logging.debug("No translation for %s in %r",message,langs)
                    status_translation['error'] = 'no_translation'
                    pass

            logging.warning("Could not find a translation in %s for %s",repr(langs),message)

        else:
            logging.debug("Message did not match any known pattern")
            self._logEvent({ 'type':'ignored', 'message':message })


    def onExit( self ):

        logging.debug("Exiting...")
        status_shutdown = { 'type':'shutdown' }
        self._logEvent(status_shutdown)

        # TODO Better use gettext in the end
        try:
            goodbye = i18n.t('Goodbye')
            if goodbye and goodbye.strip():
                text = i18n.t('all_messages',message=goodbye)
                sent = self.chatter.send(text)
                status_shutdown['answer'] = text
                status_shutdown['timestamp'] = sent
            else:
                logging.debug("Empty 'Goodbye' text : nothing was sent")
        except KeyError:
            logging.debug("No 'Goodbye' text : nothing was sent")
            pass


    def run( self ):
        """
            Starts the bot :

            1. Sends a hello message
            2. Waits for messages to translate

            Returns the execution status of the run, as a dict : { 'events':[list_of_events] }
            with list_of_events the list of input / outputs that happened, for audit purposes
        """

        self.chatter.connect()

        # TODO Better using gettext, in the end
        try:
            hello = i18n.t('Hello')
            if hello and hello.strip():
                text = i18n.t('all_messages',message=hello)
                sent = self.chatter.send(text)
                self._logEvent({ 'type':'startup', 'answer':text, 'timestamp':sent })
            else:
                logging.debug("Empty 'Hello' text : nothing was sent")
        except KeyError:
            logging.debug("No 'Hello' text : nothing was sent")
            pass

        self.registerExitHandler()
        self.chatter.start(self)
        logging.debug("Chatter loop ended")
        return self.status



def run( args=sys.argv[1:] ):

    """
        A convenient CLI to play with this bot
    """

    config = Config()

    parser = argparse.ArgumentParser(
        parents=[ BotArgsHelper().parser(), jabber_arg_parser(), SignalArgsHelper().parser() ],
        description="A bot that reacts to messages with given keywords by responding with a random translation"
        )
    # Core arguments for this bot
    parser.add_argument("--keyword", "-k", dest="keywords", action="append", help="A keyword a bot should react to (will write them into the file specified with --keywords-file)")
    parser.add_argument("--keywords-file", dest="keywords_files", action="append", help="File to load from and write keywords to")
    parser.add_argument('--locale', '-l', dest='locale', default=config.locale, help="Change default locale (e.g. 'fr_FR')")
    parser.add_argument("--languages-file", dest="languages_file", help="File to load from and write languages to")
    parser.add_argument("--languages-likely", dest="languages_likely", default=config.languages_likely, help="URI to Unicode's Likely Subtags (best language <-> country matches) in JSON format")
    parser.add_argument("--shutdown", dest="shutdown", help="Shutdown keyword regular expression pattern")
    parser.add_argument("--ibmcloud-url", dest="ibmcloud_url", help="IBM Cloud API base URL (get it from your resource https://cloud.ibm.com/resources)")
    parser.add_argument("--ibmcloud-apikey", dest="ibmcloud_apikey", help="IBM Cloud API key (get it from your resource : https://cloud.ibm.com/resources)")

    #
    # Two-pass arguments parsing
    #
    config = parse_args_2pass( parser, args, config )
    #
    # From here the config object has default options from:
    #   1. hard-coded default values
    #   2. configuration file overrides
    #   3. command line overrides
    #
    # We can check the required options that could not be checked before
    # (because required arguments may have been set from the config file and not on the command line)
    #

    # i18n + l10n
    logging.debug("Current locale : %s"%repr(locale.getlocale()))
    # e.g. if config.locale is 'en_US' we split it into : ['en', 'US'] ; dash separator is the RFC norm '-', but underscore '_' is used with Python
    lang = re.split( r'[_-]', config.locale )
    # See https://pypi.org/project/python-i18n/
    # FIXME Manually sets the locale : how come a Python library named 'i18n' doesn't take into account the Python locale by default ?
    i18n.set('locale',lang[0])
    logging.debug("i18n locale : %s"%i18n.get('locale'))
    i18n.set('filename_format', 'i18n.{locale}.{format}')    # Removing the namespace from keys is simpler for us
    i18n.set('error_on_missing_translation',True)
    i18n.load_path.append(config.config_dir)

    # These MUST be instanciated AFTER i18n has been configured !
    try:
        i18n.t('all_messages',message="")
    except:
        i18n.add_translation('all_messages',r'%{message}')
    if not config.shutdown:
        config.shutdown = i18n.t('Shutdown')

    if not config.ibmcloud_url:
        raise ValueError("Missing required parameter : --ibmcloud-url")
    if not config.ibmcloud_apikey:
        raise ValueError("Missing required parameter : --ibmcloud-apikey")

    # config.keywords is used if given
    # else, check for an existing keywords_file
    if len(config.keywords_files) == 0:
        # As a last resort, use 'keywords.json' in the config directory
        config.keywords_files = [ os.path.join(config.config_dir,'keywords.json') ]
    # Convenience check to better warn the user and allow filenames relative to config_dir
    if not config.keywords:
        found_keywords_files = []
        for keywords_file in config.keywords_files:
            relative_filename = os.path.join(config.config_dir,keywords_file)
            winners = filter_files( [keywords_file, relative_filename], should_exist=True, fallback_to=None )
            if len(winners) > 0:
                found_keywords_files = found_keywords_files + winners
        if len(found_keywords_files) > 0:
            config.keywords_files = found_keywords_files
        else:
            raise ValueError("Could not open any keywords file in %s : please generate with --keywords first or create the file indicated with --keywords-file"%repr(config.keywords_files))

    # Finds an existing languages_file
    # By default, uses 'languages.<lang>.json' or 'languages.json' in the config directory
    config.languages_file = filter_files( [
        config.languages_file,
        os.path.join( config.config_dir, "languages.%s.json"%lang[0] ),
        os.path.join( config.config_dir, 'languages.json' ) ],
        should_exist=True,
        fallback_to=1 )[0]
    # Convenience check to better warn the user
    if not config.languages_file:
        raise ValueError("Missing language file : please use only --languages-file to generate it automatically or --language for each target language")

    # Finds a "likely language" file
    config.languages_likely = filter_files([
        config.languages_likely,
        os.path.join( config.config_dir, 'likelySubtags.json' ) ],
        should_exist=True,
        fallback_to=1 )[0]

    # Creates the chat engine depending on the 'backend' parameter
    chatter = BotArgsHelper.chatter(config)

    #
    # Real start
    #

    bot = TransBot(
        keywords=config.keywords, keywords_files=config.keywords_files,
        languages_file=config.languages_file, languages_likely=config.languages_likely,
        locale=lang,
        ibmcloud_url=config.ibmcloud_url, ibmcloud_apikey=config.ibmcloud_apikey,
        shutdown_pattern=config.shutdown,
        chatter=chatter
        )
    status_args = vars(config)
    # TODO Add an option to list the fields to obfuscate (nor not)
    for k in [ 'ibmcloud_apikey', 'jabber_password' ]:
        status_args[k] = '(obfuscated)'
    status_result = bot.run()
    status = { 'args':vars(config), 'result':status_result }
    # Returns the full status to this module can be called CLI-style
    return status


# Like run(), but also prints the final status to stdout
if __name__ == '__main__':

    status = run(sys.argv[1:])
    # NOTE ensure_ascii=False + encode('utf-8').decode() is not mandatory but allows printing plain UTF-8 strings rather than \u... or \x...
    # NOTE default=repr is mandatory because some objects in the args are not serializable
    print( json.dumps(status,skipkeys=True,ensure_ascii=False,default=repr).encode('utf-8').decode(), file=sys.stdout, flush=True )
    sys.exit(0)



# -*- coding: utf-8 -*-


class Chatter:
    """
        Bot engine interface
    """

    def connect(self):
        """
            Connects / initializes the connection with the underlying protocol/network if required.
            This should always be called before any other method in order to make sure the bot is connected.
        """
        pass

    def start( self, bot ):
        """
            Waits for messages and calls the 'onMessage' method of the given Bot
        """
        pass

    def reply( self, source ):
        """
            Replies to a specific message or person
        """
        pass

    def send( self, message ):
        """
            Sends the given message using the underlying implemented chat protocol
        """
        pass

    def stop( self ):
        """
            Stops waiting for messages and exits the engine
        """
        pass



# -*- coding: utf-8 -*-

import argparse
import atexit
import logging
import os
import signal
import sys

from .console import ConsoleChatter
from .jabber import JabberChatter
from .signalcli import SignalChatter
from .stealth import StealthChatter


# There are other options but this one is the most efficient here
# https://github.com/pypa/setuptools_scm#retrieving-package-version-at-runtime
from .version import version as __version__


class Bot:
    """
        Bot foundation
    """

    def onMessage( self, message ):
        """
            Called by self.chatter whenever a message has arrived.

            message: A plain text message
            Returns nothing
        """
        pass


    def onExit( self ):
        """
            Called just before exiting ; the chatter should still be available.
            Subclass MUST call registerExitHandler for this to work !
        """
        logging.debug("Exiting...")


    def onSignal( self, sig, frame ):
        # Thanks https://stackoverflow.com/questions/23468042/the-invocation-of-signal-handler-and-atexit-handler-in-python
        logging.debug("Got signal %s %s",sig,repr(frame))
        sys.exit(0)


    def registerExitHandler( self ):
        # Registers exit handlers to properly say goodbye
        atexit.register(self.onExit)
        # TODO This list does not work on Windows
        for sig in [signal.SIGINT, signal.SIGTERM, signal.SIGHUP ]:
            signal.signal(sig, self.onSignal)

    def run( self ):
        """
            Starts the bot
        """
        pass



class ArgsHelper:

    """
        Command-line parsing helper for bot-generic options
    """

    def __init__(self):

        # Default configuration (some defaults still need to be set up after command line has been parsed)
        self.__dict__.update({
            'backend': "console",
            'config_file': "config.yml",
            'config_dir': os.getcwd(),
            'input_file': sys.stdin,
            'stealth': False,
            'verbosity': "WARNING",
            })


    def parser(self):
        """
            Returns a parent parser for common bot arguments
        """

        parser = argparse.ArgumentParser(add_help=False)

        # Bootstrap options
        parser.add_argument("--config-file", "-c", "--config", dest="config_file", default=self.config_file, help="YAML configuration file.")
        parser.add_argument("--config-dir", "-C", dest="config_dir", default=self.config_dir, help="Directory where to find configuration files by default.")
        parser.add_argument('--verbosity', '-v', dest='verbosity', default=self.verbosity, help="Log level")
        # Chatter-generic arguments
        parser.add_argument("--backend", "-b", dest="backend", choices=['console','jabber','signal'], default=self.backend, help="Chat backend to use")
        parser.add_argument("--input-file", "-i", dest="input_file", default=self.input_file, help="File to read messages from (one per line)")
        parser.add_argument('--username', '-U', dest='username', help="Sender's ID (a phone number for Signal, a Jabber Identifier (JID) aka. username for Jabber/XMPP")
        parser.add_argument('--recipient', '-r', '--receiver', dest='recipients', default=[], action='append', help="Recipient's ID (e.g. '+12345678901' for Signal / JabberID (Receiver address) to send the message to)")
        parser.add_argument('--stealth', dest='stealth', action="store_true", default=self.stealth, help="Activate stealth mode on any chosen chatter")
        # Misc. options
        parser.add_argument("--debug", "-d", action="store_true", dest='debug', default=False, help="Activate debug logs (overrides --verbosity)")
        # Needs the .git metadata or some variables to be able to return the version, otherwise throws an error
        # See https://github.com/pypa/setuptools_scm
        parser.add_argument("--version", "-V", action="version", version=__version__)

        return parser


    def jabber_chatter( args ):
        """
            Builds a JabberChatter from Namespace argument 'args'.
            Sets its data directory to <config_dir>/.omemo
        """

        username = args.jabber_username if args.jabber_username else args.username
        if not username:
            raise ValueError("Missing --jabber-username")
        if not args.jabber_password:
            raise ValueError("Missing --jabber-password")
        recipients = args.jabber_recipients + args.recipients
        if len(recipients)==0:
            raise ValueError("Missing --jabber-recipient")
        # TODO allow multiple recipients
        return JabberChatter(
            jid=username,
            password=args.jabber_password,
            recipient=recipients[0],
            data_dir=os.path.join(args.config_dir,".omemo")
            )


    def signal_chatter( args ):
        """
            Builds a SignalChatter from Namespace argument 'args'
        """

        if not args.signal_cli:
            raise ValueError("Could not find the 'signal-cli' command in PATH and no --signal-cli given")
        username = args.signal_username if args.signal_username else args.username
        if not username:
            raise ValueError("Missing --signal-username")
        recipients = args.signal_recipients + args.recipients
        if len(recipients)==0 and not args.signal_group:
            raise ValueError("Either --signal-recipient or --signal-group must be provided")
        # TODO allow multiple recipients
        return SignalChatter(
            username=username,
            recipient=recipients[0],
            group=args.signal_group,
            signal_cli=args.signal_cli,
            stealth=args.signal_stealth
            )
        # TODO  :timeout=args.timeout


    def chatter( args ):
        """
            Builds the Chatter corresponding to the given parsed command-line arguments
            args: command-line arguments as a Namespace (see argparse)
        """

        if args.backend == 'jabber':
            logging.debug("Jabber/XMPP backend selected")
            chatter = ArgsHelper.jabber_chatter(args)

        elif args.backend == 'signal':
            logging.debug("Signal backend selected")
            chatter = ArgsHelper.signal_chatter(args)

        # By default (or if backend == "console"), will read from stdin or a given file and output to console
        else:
            logging.debug("Console backend selected")
            chatter = ConsoleChatter(args.input_file)

        if args.stealth:
            chatter = StealthChatter(chatter)

        return chatter



# -*- coding: utf-8 -*-

import logging
import sys

from .chatter import Chatter


class StealthChatter(Chatter):
    """
        Wraps a bot engine and prints messages rather than sending them
    """

    def __init__( self, chatter, output=sys.stdout ):
        self.chatter = chatter
        self.output = output

    def start( self, bot ):
        logging.debug("Stealth mode : will not send message")
        self.chatter.start( bot )

    def send( self, message ):
        print( message, file=self.output, flush=True )

    def stop( self ):
        self.chatter.stop()



# -*- coding: utf-8 -*-

import argparse
import asyncio
import logging
import os
import time

from slixmpp import ClientXMPP, JID
from slixmpp.exceptions import IqTimeout, IqError
from slixmpp.stanza import Message
import slixmpp_omemo
from slixmpp_omemo import PluginCouldNotLoad, MissingOwnKey, EncryptionPrepareException
from slixmpp_omemo import UndecidedException, UntrustedException, NoAvailableSession
from omemo.exceptions import MissingBundleException

# Own classes
from .chatter import Chatter
from .helpers import *

log = logging.getLogger(__name__)



class SliXmppClient(ClientXMPP):

    """
        This generic XMPP client is able to send & receive plain & OMEMO-encrypted messages.

        Code is mostly taken from https://lab.louiz.org/poezio/slixmpp-omemo/-/blob/master/examples/echo_client.py
    """

    eme_ns = 'eu.siacs.conversations.axolotl'

    def __init__(self, jid, password, message_handler, data_dir):

        """
            jid, password : valid account to send and receive messages
            message_handler : a Callable( original_message:Message, decrypted_body )
        """

        ClientXMPP.__init__(self, jid, password)

        self.add_event_handler("session_start", self.session_start)
        self.add_event_handler("message", self.message)

        self.register_plugin('xep_0030') # Service Discovery
        self.register_plugin('xep_0199') # XMPP Ping
        self.register_plugin('xep_0380') # Explicit Message Encryption

        if not os.path.exists(data_dir):
            os.makedirs(data_dir)
        try:
            self.register_plugin(
                'xep_0384',
                {
                    'data_dir': data_dir,
                },
                module=slixmpp_omemo,
            ) # OMEMO
        except (PluginCouldNotLoad,):
            log.exception('And error occured when loading the omemo plugin.')
            sys.exit(1)

        self.message_handler = message_handler


    def session_start(self, event):
        self.send_presence()
        self.get_roster()

        # Most get_*/set_* methods from plugins use Iq stanzas, which
        # can generate IqError and IqTimeout exceptions
        #
        # try:
        #     self.get_roster()
        # except IqError as err:
        #     log.error('There was an error getting the roster')
        #     log.error(err.iq['error']['condition'])
        #     self.disconnect()
        # except IqTimeout:
        #     log.error('Server is taking too long to respond')
        #     self.disconnect()


    async def message(self, msg: Message, allow_untrusted: bool = False) -> None:
        """
        Process incoming message stanzas. Be aware that this also
        includes MUC messages and error messages. It is usually
        a good idea to check the messages's type before processing
        or sending replies.

        Arguments:
            msg -- The received message stanza. See the documentation
                   for stanza objects and the Message stanza to see
                   how it may be used.
        """

        # Sample encrypted message :
        #
        # <message xml:lang="en" to="bot9cd51f1a@conversations.im/1655465886336291765177990819" from="bot649ad4ad@conversations.im/Conversations.7Z6J" type="chat" id="26a30844-37d0-4f27-8daa-fd2107f5e706"><archived xmlns="urn:xmpp:mam:tmp" by="bot9cd51f1a@conversations.im" id="1590067423912722"/><stanza-id xmlns="urn:xmpp:sid:0" by="bot9cd51f1a@conversations.im" id="1590067423912722"/>
        #     <encrypted xmlns="eu.siacs.conversations.axolotl">
        #         <header sid="307701646">
        #             <key rid="45781751">MwohBadGcbUIiKsYACIw8UjlZHsOEf79+fjBM44O9bM1YatXKRKaIYuaIkajsedDkIK906Srxk2N5B7jh8EozEAFKpfeqZg//Hqrin6wOGpX+TZ5kUJziXIALUaRs59D3J0=</key>
        #             <key rid="1929813965">MwohBWTCE+eWLNmdgTy/gyEAAYACIwzz9+B/UFiaCu+5rcHmh3tyQ/GgBhVa+mk81YkQErXpjCpAPyWbKVJn2TH1dXH4Yj7VYYis0HDQ7r28ZDcMMXoxFcp2VNO9l7S23wI=</key>
        #             <iv>5eM9IHpWSbKfLJj6</iv>
        #         </header>
        #         <payload>pX7D+54c</payload>
        #     </encrypted><request xmlns="urn:xmpp:receipts"/><markable xmlns="urn:xmpp:chat-markers:0"/><origin-id xmlns="urn:xmpp:sid:0" id="26a30844-37d0-4f27-8daa-fd2107f5e706"/><store xmlns="urn:xmpp:hints"/><encryption xmlns="urn:xmpp:eme:0" name="OMEMO" namespace="eu.siacs.conversations.axolotl"/>
        #     <body>I sent you an OMEMO encrypted message but your client doesn’t seem to support that. Find more information on https://conversations.im/omemo</body>
        # </message>

        # Sample plain message :
        # <message xml:lang="en" to="bot9cd51f1a@conversations.im/11834511835037473566179797763" from="bot649ad4ad@conversations.im/Conversations.7Z6J" type="chat" id="5677cbdf-11d2-4aed-889e-0fb3e850a390"><archived xmlns="urn:xmpp:mam:tmp" by="bot9cd51f1a@conversations.im" id="1590095998371956"/><stanza-id xmlns="urn:xmpp:sid:0" by="bot9cd51f1a@conversations.im" id="1590095998371956"/><request xmlns="urn:xmpp:receipts"/><markable xmlns="urn:xmpp:chat-markers:0"/><origin-id xmlns="urn:xmpp:sid:0" id="5677cbdf-11d2-4aed-889e-0fb3e850a390"/><active xmlns="http://jabber.org/protocol/chatstates"/>
        #     <body>My message</body>
        # </message>

        log.debug("XMPP message received : %r",msg)

        # TODO ? with xmppy I used to allow the following types : ["message","chat","normal",None]
        if msg['type'] not in ('chat', 'normal'):
            log.debug("Discarding message of type %r",msg['type'])
            return None

        if not self['xep_0384'].is_encrypted(msg):
            log.debug('This message was not encrypted')
            self.message_handler(msg,msg['body'])
            return None

        try:
            mfrom = msg['from']
            encrypted = msg['omemo_encrypted']
            body = self['xep_0384'].decrypt_message(encrypted, mfrom, allow_untrusted)
            # TODO Is it always UTF-8-encoded ?
            self.message_handler(msg,body.decode("utf8"))
            return None
        except (MissingOwnKey,):
            # The message is missing our own key, it was not encrypted for
            # us, and we can't decrypt it.
            log.exception('I can\'t decrypt this message as it is not encrypted for me : %r',msg)
            return None
        except (NoAvailableSession,):
            # We received a message from that contained a session that we
            # don't know about (deleted session storage, etc.). We can't
            # decrypt the message, and it's going to be lost.
            # Here, as we need to initiate a new encrypted session, it is
            # best if we send an encrypted message directly. XXX: Is it
            # where we talk about self-healing messages?
            log.exception('I can\'t decrypt this message as it uses an encrypted session I don\'t know about : %r',msg)
            return None
        except (UndecidedException, UntrustedException) as exn:
            # We received a message from an untrusted device. We can
            # choose to decrypt the message nonetheless, with the
            # `allow_untrusted` flag on the `decrypt_message` call, which
            # we will do here. This is only possible for decryption,
            # encryption will require us to decide if we trust the device
            # or not. Clients _should_ indicate that the message was not
            # trusted, or in undecided state, if they decide to decrypt it
            # anyway.
            log.exception("Your device '%s' is not in my trusted devices.", exn.device)
            # We resend, setting the `allow_untrusted` parameter to True.
            await self.message(msg, allow_untrusted=True)
            return None
        except (EncryptionPrepareException,):
            # Slixmpp tried its best, but there were errors it couldn't
            # resolve. At this point you should have seen other exceptions
            # and given a chance to resolve them already.
            log.exception('I was not able to decrypt the message : %r',msg)
            return None
        except (Exception,):
            log.exception('An error occured while attempting decryption')
            raise

        return None


    async def plain_send(self, body, receiver, type='chat'):
        """
        Helper to send messages
        """

        msg = self.make_message(mto=receiver, mtype=type)
        msg['body'] = body
        return msg.send()


    async def plain_reply(self, original_msg, body):
        """
        Helper to reply to messages
        """

        return self.plain_send( body, original_msg['from'], original_msg['type'] )


    async def encrypted_send(self, body, recipient, type='chat'):
        """Helper to send encrypted messages"""

        msg = self.make_message(mto=recipient, mtype=type)
        msg['eme']['namespace'] = self.eme_ns
        msg['eme']['name'] = self['xep_0380'].mechanisms[self.eme_ns]

        expect_problems = {}  # type: Optional[Dict[JID, List[int]]]

        while True:
            try:
                # `encrypt_message` excepts the plaintext to be sent, a list of
                # bare JIDs to encrypt to, and optionally a dict of problems to
                # expect per bare JID.
                #
                # Note that this function returns an `<encrypted/>` object,
                # and not a full Message stanza. This combined with the
                # `recipients` parameter that requires for a list of JIDs,
                # allows you to encrypt for 1:1 as well as groupchats (MUC).
                #
                # `expect_problems`: See EncryptionPrepareException handling.
                recipients = [JID(recipient)]
                encrypt = await self['xep_0384'].encrypt_message(body, recipients, expect_problems)
                msg.append(encrypt)
                return msg.send()
            except UndecidedException as exn:
                # The library prevents us from sending a message to an
                # untrusted/undecided barejid, so we need to make a decision here.
                # This is where you prompt your user to ask what to do. In
                # this bot we will automatically trust undecided recipients.
                self['xep_0384'].trust(exn.bare_jid, exn.device, exn.ik)
            # TODO: catch NoEligibleDevicesException
            except EncryptionPrepareException as exn:
                # This exception is being raised when the library has tried
                # all it could and doesn't know what to do anymore. It
                # contains a list of exceptions that the user must resolve, or
                # explicitely ignore via `expect_problems`.
                # TODO: We might need to bail out here if errors are the same?
                for error in exn.errors:
                    if isinstance(error, MissingBundleException):
                        # We choose to ignore MissingBundleException. It seems
                        # to be somewhat accepted that it's better not to
                        # encrypt for a device if it has problems and encrypt
                        # for the rest, rather than error out. The "faulty"
                        # device won't be able to decrypt and should display a
                        # generic message. The receiving end-user at this
                        # point can bring up the issue if it happens.
                        log.warning('Could not find keys for device "%d" of recipient "%s". Skipping.', error.device, error.bare_jid)
                        jid = JID(error.bare_jid)
                        device_list = expect_problems.setdefault(jid, [])
                        device_list.append(error.device)
            except (IqError, IqTimeout) as exn:
                log.exception('An error occured while fetching information on %r', recipient)
                return None
            except Exception as exn:
                log.exception('An error occured while attempting to encrypt to %r', recipient)
                raise

        return None


    async def encrypted_reply(self, original_msg, body):
        """Helper to reply with encrypted messages"""

        log.debug("Replying to %r",original_msg)
        return self.encrypted_send( body, recipient=original_msg['from'], type=original_msg['type'] )



class JabberChatter(Chatter):

    """
        Sends and receives messages with XMPP (a.k.a. Jabber).

        It implements nicobot.Chatter by wrapping an internal slixmpp.ClientXMPP instance.
    """

    def __init__( self, jid, password, recipient, data_dir ):

        self.recipient = recipient
        self.xmpp = SliXmppClient( jid, password, message_handler=self.on_xmpp_message, data_dir=data_dir )

    def on_xmpp_message( self, original_message, decrypted_body ):
        """
            Called by the internal xmpp client when a message has arrived.

            original_message: The received Message instance
            decrypted_body: either the given body if it was plain text or the OMEMO-decrypted one (always a string)
        """

        log.log(TRACE,"<<< %r",original_message)
        log.debug("<<< %r",decrypted_body)
        self.bot.onMessage(decrypted_body)

    def connect(self):

        log.debug("Connecting...")
        # Connects and waits for the connection to be established
        # See https://slixmpp.readthedocs.io/using_asyncio.html
        self.xmpp.connected_event = asyncio.Event()
        callback = lambda _: self.xmpp.connected_event.set()
        self.xmpp.add_event_handler('session_start', callback)
        self.xmpp.connect()
        # TODO use asyncio.run() in latest python
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.xmpp.connected_event.wait())
        log.debug("Connected.")

    def start( self, bot ):
        """
            Waits for messages and calls the 'onMessage' method of the given Bot
        """
        self.bot = bot
        # do some other stuff before running the event loop, e.g.
        # loop.run_until_complete(httpserver.init())
        self.xmpp.process(forever=False)
        # FIXME Following error when exiting :
        # Task was destroyed but it is pending! task: <Task pending coro=<XMLStream.run_filters() running at /home/./.local/lib/python3.6/site-packages/slixmpp/xmlstream/xmlstream.py:972> wait_for=<Future pending cb=[<TaskWakeupMethWrapper object at 0x7fa91adc0fd8>()]>>

    def send( self, message ):
        """
            Sends the given message using the underlying implemented chat protocol
        """
        log.debug(">>> %s",message)
        # TODO use asyncio.make_task() in latest python
        asyncio.ensure_future( self.xmpp.encrypted_send( body=message, recipient=self.recipient ) )

    def stop( self ):
        """
            Stops waiting for messages and exits the engine
        """
        self.xmpp.disconnect()



def arg_parser():
    """
        Returns a parent parser for jabber-specific arguments
    """

    parser = argparse.ArgumentParser(add_help=False)

    # Jabber-specific arguments
    parser.add_argument('--jabber-username', '--jabberid', '--jid', dest='jabber_username', help="Username when using the Jabber/XMPP backend (overrides --username)")
    parser.add_argument('--jabber-recipient', dest='jabber_recipients', action='append', default=[], help="Recipient when using the Jabber/XMPP backend (overrides --recipient)")
    parser.add_argument('--jabber-password', dest='jabber_password', help="Senders's password")

    return parser



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import logging
import sys
import os
import shutil
import json
import i18n
import re
import locale
import requests
import random
# Provides an easy way to get the unicode sequence for country flags
import flag
import yaml
import urllib.request

# Own classes
from .helpers import *
from .bot import Bot
from .bot import ArgsHelper as BotArgsHelper
from .console import ConsoleChatter
from .jabber import JabberChatter
from .jabber import arg_parser as jabber_arg_parser
from .signalcli import SignalChatter
from .signalcli import ArgsHelper as SignalArgsHelper
from .stealth import StealthChatter


# Default configuration (some defaults still need to be set up after command line has been parsed)
class Config:

    def __init__(self):
        self.__dict__.update({
            'backend': "console",
            'config_file': "config.yml",
            'config_dir': os.getcwd(),
            'input_file': sys.stdin,
            'max_count': -1,
            'patterns': [],
            'stealth': False,
            'timeout': None,
            'verbosity': "WARNING",
            })



class AskBot(Bot):
    """
        Sends a message and reads the answer.
        Can be configured with retries, pattern matching, ...

        patterns : a list of 2-element list/tuple as [name,pattern]
    """

    def __init__( self, chatter, message, patterns=[], max_count=-1 ):

        # TODO Implement a global session timeout after which the bot exits
        self.status = {
            'max_count': False,
            'events': [],
        }
        self.responses_count = 0

        self.chatter = chatter
        self.message = message
        self.max_count = max_count
        self.patterns = []
        for pattern in patterns:
            self.patterns.append({ 'name':pattern[0], 'pattern':re.compile(pattern[1]) })


    def onMessage( self, message ):
        """
            Called by self.chatter whenever a message has arrived.

            message: A plain text message
            Returns the full status with exit conditions
        """

        status_message = { 'message':message, 'matched_patterns':[] }
        self.status['events'].append(status_message)

        self.responses_count = self.responses_count + 1
        logging.info("<<< %s", message)

        # If we reached the last message or if we exceeded it (possible if we received several answers in a batch)
        if self.max_count>0 and self.responses_count >= self.max_count:
            logging.debug("Max amount of messages reached")
            self.status['max_count'] = True

        # Another way to quit : pattern matching
        matched = status_message['matched_patterns']
        for p in self.patterns:
            name = p['name']
            pattern = p['pattern']
            if pattern.search(message):
                logging.debug("Pattern '%s' matched",name)
                matched.append(name)

        # Check if any exit condition is met to notify the underlying chatter engine
        if self.status['max_count'] or len(matched) > 0:
            logging.debug("At least one pattern matched : exiting...")
            self.chatter.stop()


    def run( self ):
        """
            Starts the bot :

            1. Sends the given message(s)
            2. Reads and print maximum 'attempts' messages

            Returns the execution status of this bot
        """

        logging.debug("Bot ready.")
        self.registerExitHandler()

        self.chatter.connect()
        # FIXME Sometimes the message is not received by the recipient (but the logs show it's sent ?)
        if self.message:
            self.chatter.send(self.message)

        # Blocks on this line until the bot exits
        logging.debug("Bot reading answer...")
        self.chatter.start(self)

        logging.debug("Bot done.")
        return self.status



def run( args=sys.argv[1:] ):

    """
        A convenient CLI to play with this bot.
    """

    # config will be the final, merged configuration
    config = Config()

    parser = argparse.ArgumentParser(
        parents=[ BotArgsHelper().parser(), jabber_arg_parser(), SignalArgsHelper().parser() ],
        description='Sends a XMPP message and reads the answer',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter )
    # Core options for this bot
    parser.add_argument('--max-count', dest='max_count', type=int, default=config.max_count, help="Read this maximum number of responses before exiting")
    parser.add_argument('--message', '-m', dest='message', help="Message to send. If missing, will read from --input-file")
    parser.add_argument('--message-file', '-f', dest='message_file', type=argparse.FileType('r'), default=sys.stdin, help="File with the message to send. If missing, will be read from standard input")
    parser.add_argument('--pattern', '-p', dest='patterns', action='append', nargs=2, help="Exits with status 0 whenever a message matches this pattern ; otherwise with status 1")
    parser.add_argument('--timeout', '-t', dest='timeout', type=int, default=config.timeout, help="How much time t wait for an answer before quiting (in seconds)")

    #
    # Two-pass arguments parsing
    #
    config = parse_args_2pass( parser, args, config )
    #
    # From here the config object has default options from:
    #   1. hard-coded default values
    #   2. configuration file overrides
    #   3. command line overrides
    #
    # We can now check the required options that could not be checked before
    # (because required arguments may have been set from the config file and not on the command line)
    #

    # Creates the chat engine depending on the 'backend' parameter
    chatter = BotArgsHelper.chatter(config)

    #
    # Real start
    #

    bot = AskBot(
        chatter=chatter,
        message=config.message,
        patterns=config.patterns,
        max_count=config.max_count
        )
    status_args = vars(config)
    # TODO Add an option to list the fields to obfuscate (nor not)
    for k in [ 'jabber_password' ]:
        status_args[k] = '(obfuscated)'
    status_result = bot.run()
    status = { 'args':vars(config), 'result':status_result }
    # Returns the full status to this module can be called CLI-style
    return status


# Like run(), but also prints the final status to stdout
if __name__ == '__main__':

    status = run(sys.argv[1:])
    # NOTE ensure_ascii=False + encode('utf-8').decode() is not mandatory but allows printing plain UTF-8 strings rather than \u... or \x...
    # NOTE default=repr is mandatory because some objects in the args are not serializable
    print( json.dumps(status,skipkeys=True,ensure_ascii=False,default=repr).encode('utf-8').decode(), file=sys.stdout, flush=True )
    sys.exit(0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest

from nicobot.askbot import AskBot
from nicobot.console import ConsoleChatter


class TestAskbot(unittest.TestCase):

    def test_end_on_max_count(self):
        bot = AskBot(
            chatter = ConsoleChatter( input=["One","Two","Three"] ),
            message = "Let's count !",
            patterns=[
                [ "four", r'(?i)\bfour\b' ],
            ],
            max_count=3
            )
        result = bot.run()
        expected = {'max_count': True, 'events': [
            {'message': 'One', 'matched_patterns': []},
            {'message': 'Two', 'matched_patterns': []},
            {'message': 'Three', 'matched_patterns': []}]
            }
        self.assertEqual(expected,result)


    def test_end_on_yes(self):
        bot = AskBot(
            chatter = ConsoleChatter( input=["Yes !"] ),
            message = "ça va ?",
            patterns=[
                [ "yes", r'(?i)\b(yes|ok)\b' ],
                [ "no", r'(?i)\bno\b' ],
                [ "cancel", r'(?i)\b(cancel|abort)\b' ]
            ],
            max_count=-1
            )
        result = bot.run()
        expected = {'max_count': False, 'events': [{'message': 'Yes !', 'matched_patterns': ['yes']}]}
        self.assertEqual(expected,result)


    def test_dont_end_on_coucou(self):
        bot = AskBot(
            chatter = ConsoleChatter( input=["Coucou","Yes !"] ),
            message = "ça va ?",
            patterns=[
                [ "yes", r'(?i)\b(yes|ok)\b' ],
                [ "no", r'(?i)\bno\b' ],
                [ "cancel", r'(?i)\b(cancel|abort)\b' ]
            ],
            max_count=-1
            )
        result = bot.run()
        expected = {'max_count': False, 'events': [{'message': 'Coucou', 'matched_patterns': []}, {'message': 'Yes !', 'matched_patterns': ['yes']}]}
        self.assertEqual(expected,result)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import atexit
import i18n
import json
import locale
import logging
import os
import re
import shutil
import signal
import subprocess
import sys
import time

from .chatter import Chatter
from .helpers import *


# Generic timeout for signal-cli commands to return (actually only 'send' because 'receive' uses its own timeout)
SEND_TIMEOUT = 30
# Custom timeout to pass to signal-cli when receiving messages (negative values disable timeout)
RECEIVE_TIMEOUT = 5


class SignalChatter(Chatter):
    """
        A signal bot relying on signal-cli
    """

    def __init__( self, username, recipient=None, group=None, signal_cli=shutil.which("signal-cli"), stealth=False, send_timeout=SEND_TIMEOUT, receive_timeout=RECEIVE_TIMEOUT ):

        """
            stealth: if True, will connect and listen to messages but instead of sending answers, will print them
        """

        if not username or not signal_cli:
            raise ValueError("username and signal_cli must be provided")
        if not recipient and not group:
            raise ValueError("Either a recipient or a group must be given")
        if recipient and group:
            raise ValueError("Only one of recipient and group may be given")

        self.username = username
        self.recipient = recipient
        self.group = group
        self.signal_cli = signal_cli
        self.stealth = stealth
        if self.stealth:
            logging.debug("Stealth mode : will not send message")
        self.send_timeout = send_timeout
        self.receive_timeout = receive_timeout

        # Properties set elsewhere
        self.startTime = None
        # If True, will terminate the main loop
        self.shutdown = False
        self.bot = None


    def start( self, bot ):

        self.bot = bot
        # Timestamp in Signal messages is a number of milliseconds since the epoch
        # See https://github.com/signalapp/libsignal-service-java/blob/a88d6a65330ab311079e198dedd25605b1aecc5f/java/src/main/java/org/whispersystems/signalservice/api/messages/SignalServiceDataMessage.java#L344
        self.startTime = time.time() * 1000
        logging.debug("Started at %f",self.startTime)

        while not self.shutdown:
            self.filterMessages( self.receiveMessages() )


    def send( self,  message ):

        cmd = [ self.signal_cli, "-u", self.username, "send", "-m", message ]
        if self.recipient:
            cmd = cmd + [ self.recipient ]
        elif self.group:
            cmd = cmd + [ "-g", self.group ]

        # throws an error in case of status <> 0
        logging.debug(cmd)
        if not self.stealth:
            proc = subprocess.run( cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, check=True, timeout=self.send_timeout )
            sent = proc.stdout
            logging.debug("Sent message : %s"%repr(sent))
        logging.debug( ">>> %s" % message )


    def reply( self, source ):
        # TODO
        pass


    def stop( self ):

        logging.debug("Stopping...")
        self.shutdown = True


    def receiveMessages( self, timeout=None, input=None ):
        """
            timeout: uses self.receive_timeout by default ; negative values disable timeout
        """

        if not timeout:
            timeout = self.receive_timeout

        cmd = [ self.signal_cli, "-u", self.username, "receive", "--json" ]
        if timeout:
            cmd = cmd + [ "-t", str(timeout) ]

        if not input:
            # This log can be very verbose and unuseful when reading empty responses every few seconds
            logging.log(TRACE,cmd)
            proc = subprocess.Popen( cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE )
            input = proc.stdout
        events = []
        for bline in iter(input.readline, b''):
            logging.debug("Read line : %s" % bline)
            # try:
            #     line = bline.decode()
            # except (UnicodeDecodeError, AttributeError):
            #     line = bline
            events = events + [json.loads(bline.rstrip())]

        return events


    def filterMessages( self, events ):

        for event in events:
            logging.debug("Filtering message : %s" % repr(event))
            envelope = event['envelope']
            if envelope['timestamp'] > self.startTime:
                # TODO This test prevents sending and receiving with the same number
                # See https://github.com/nicolabs/nicobot/issues/34
                if envelope['dataMessage']:
                    dataMessage = envelope['dataMessage']
                    if dataMessage['message']:
                        message = event['envelope']['dataMessage']['message']
                        if self.recipient:
                            if envelope['source'] == self.recipient:
                                logging.debug("<<< %s" % message)
                                self.bot.onMessage(message)
                            else:
                                logging.debug("Discarding message not from recipient %s"%self.recipient)
                        elif self.group:
                            if dataMessage['groupInfo'] and dataMessage['groupInfo']['groupId']:
                                logging.debug("<<< %s" % message)
                                self.bot.onMessage(message)
                            else:
                                logging.debug("Discarding message not from group %s" % self.group)
                        else:
                            raise ValueError("Neither a recipient nor a group was configured : we should not be here")
                    else:
                        logging.debug("Discarding message without text")
                else:
                    logging.debug("Discarding message without data")
            else:
                logging.debug("Discarding message that was sent before I started")



class ArgsHelper:

    """
        Command-line parsing helper for Signal-specific options
    """

    def __init__(self):

        # Default configuration (some defaults still need to be set up after command line has been parsed)
        self.__dict__.update({
            'signal_cli': shutil.which("signal-cli"),
            'signal_stealth': False,
            })

    def parser(self):
        """
            Returns a parent parser for Signal-specific arguments
        """

        parser = argparse.ArgumentParser(add_help=False)

        # Signal-specific arguments
        parser.add_argument('--signal-cli', dest='signal_cli', default=self.signal_cli, help="Path to `signal-cli` if not in PATH")
        parser.add_argument('--signal-username', dest='signal_username', help="Username when using the Signal backend (overrides --username)")
        parser.add_argument('--signal-group', dest='signal_group', help="Group's ID (for Signal : a base64 string (e.g. 'mPC9JNVoKDGz0YeZMsbL1Q==')")
        parser.add_argument('--signal-recipient', dest='signal_recipients', action='append', default=[], help="Recipient when using the Signal backend (overrides --recipient)")
        parser.add_argument('--signal-stealth', dest='signal_stealth', action="store_true", default=self.signal_stealth, help="Activate Signal chatter's specific stealth mode")

        return parser



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import setuptools
#from setuptools_scm import get_version
import re


with open("README.md", "r") as fh:
    long_description = fh.read()


def local_scheme(version):
    return ""

"""
    Very specific Docker tag naming for this project
    This could have been done in Sh, Makefile, ... but here it's integrated
    with the setup script.

    See https://dankeder.com/posts/adding-custom-commands-to-setup-py/
    What variables are available from Githu Actions : https://docs.github.com/en/actions/reference/context-and-expression-syntax-for-github-actions#github-context
"""
class DockerTagsCommand(setuptools.Command):

    description = "Prints a tag list for the given 'base tag'"
    user_options = [
          ('image=', 'i', 'Image name (defaults to nicolabs/nicobot)'),
          ('variant=', 'v', 'Image variant / base tag : debian|debian-signal|alpine'),
          ('branch=', 'b', 'The git ref as <branch name> or refs/heads/<branch name>'),
          ('tag=', 't', 'The git ref as <tag name> or refs/tags/<tag name>'),
          ('ref=', 'r', 'The git ref as refs/tags|heads/<tag or branch name>'),
          ('sep=', 's', 'A string to separate each tag in the output (defaults to a new line)'),
      ]

    def initialize_options(self):
        self.image = 'nicolabs/nicobot'
        self.variant = None
        self.branch = None
        self.tag = None
        self.ref = None
        self.sep = '\n'

    def finalize_options(self):
        """
            ref is the backup value for either tag or branch
            final tag & branch values are normalized anyway
        """

        stag = self.tag
        if not stag and self.ref:
            stag = self.ref
        try:
            # If semver, we don't retain the 'v' prefix
            self.tag = re.search('refs/tags/v?(.*)', stag, flags=re.IGNORECASE).group(1)
        except:
            pass

        sbranch = self.branch
        if not sbranch and self.ref:
            sbranch = self.ref
        try:
            self.branch = re.search('refs/heads/(.*)', sbranch, flags=re.IGNORECASE).group(1)
        except:
            pass

    def run(self):
        # # TODO How to get the actual configuration as set up by setuptools ?
        # version = get_version(local_scheme='no-local-version')
        tags = []
        if self.tag:
            # When git-tagged, the variant name alone means : the 'latest' for this variant
            tags = [ "%s:%s" % (self.image,self.variant) ]
            # It is also tagged with the version
            tags = tags + [ "%s:%s-%s" % (self.image,self.variant,self.tag) ]
            # Only debian-signal is tagged with 'latest' additionaly
            if self.variant == "debian-signal":
                tags = tags + [ "%s:latest" % (self.image) ]
        elif self.branch:
            # All non git-tagged commits overwrite the 'dev' tag (only one is useful currently)
            tags = tags + [ "%s:%s-dev" % (self.image,self.variant) ]
        print( self.sep.join(tags) )


setuptools.setup(
    # See https://packaging.python.org/tutorials/packaging-projects/
    name="nicobot",
    author="nicobo",
    author_email="nicobo@users.noreply.github.com",
    description="A collection of 🤟 cool 🤟 chat bots",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nicolabs/nicobot",
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha',
        'Topic :: Communications :: Chat',
        'Topic :: Internet :: XMPP',
        'Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator',
        'Topic :: Software Development :: Internationalization'
    ],
    # PyYAML requires Python 3.5+
    python_requires='>=3.5',
    # TODO This duplicates requirements-build.txt ?
    setup_requires=['setuptools-scm'],
    # TODO This duplicates requirements-runtime.txt
    # Is runnning setup.py enough to replace pip install -r requirements-runtime.txt ?
    install_requires=[
        ##### Requirements for signalcli #####
        'python-i18n',
        ###### Requirements for transbot #####
        'python-i18n',
        # https://requests.readthedocs.io/en/master/
        'requests',
        # https://github.com/cvzi/flag
        'emoji-country-flag',
        # https://pyyaml.org/wiki/PyYAMLDocumentation
        'pyyaml',
        ###### Requirements for jabber #####
        'slixmpp-omemo',
    ],
    entry_points={
        'console_scripts': [
            'askbot=nicobot.askbot:run',
            'transbot=nicobot.transbot:run',
        ],
    },
    # Extracts version from SCM ; https://github.com/pypa/setuptools_scm/
    use_scm_version = {
        "write_to": "nicobot/version.py",
        # Only enable to upload local versions to test repo
        # See https://mixstersite.wordpress.com/2019/12/31/setuptools-with-testpypi-error-invalid-version-pep-440/
        # and https://github.com/pypa/setuptools_scm#user-content-version-number-construction
        "local_scheme": 'no-local-version',
    },
    cmdclass = {
        'docker_tags': DockerTagsCommand,
    },
)



# -*- coding: utf-8 -*-

"""
    Helper functions
"""

import logging
import os
import sys
import yaml


# Adds a log level finer than DEBUG
TRACE = 5
logging.addLevelName(TRACE,'TRACE')


def configure_logging( level=None, debug=None ):
    """
        Sets default logging preferences for this module

        if debug=True, overrides level with DEBUG
    """

    if debug:
        logLevel = logging.DEBUG
    else:
        try:
            # Before Python 3.4 and back since 3.4.2 we can simply pass a level name rather than a numeric value (Yes !)
            # Otherwise manually parsing textual log levels was not clean IMHO anyway : https://docs.python.org/2/howto/logging.html#logging-to-a-file
            logLevel = logging.getLevelName(level.upper())
        except ValueError:
        	raise ValueError('Invalid log level: %s' % level)

    # Logs are output to stderr ; stdout is reserved to print the answer(s)
    logging.basicConfig(level=logLevel, stream=sys.stderr, format='%(asctime)s\t%(levelname)s\t%(message)s')


def filter_files( files, should_exist=False, fallback_to=None ):
    """
        files: a list of filenames / open files to filter
        should_exist: filters out non-existing files
        fallback_to: if the result would be an empty list, add the entry with this index in the result ; ignored if None
        Returns : a list with only the files that passed the filters
    """

    found = []
    for file in files:
        if should_exist:
            try:
                with open(file,'r') as f:
                    pass
            except:
                continue
        found = found + [file]

    if len(found) == 0 and fallback_to is not None:
        return files[fallback_to:fallback_to+1]

    return found


def parse_args_2pass( parser, args, config ):
    """
        Wrapper around argparse's ArgumentParser.parse_args that makes two passes :

        1. the first one to identify a configuration file and load defaults from it
        2. the second one to parse other command-line parameters falling back to defaults from the config file for missing parameters

        It also takes care of logging configuration, which is considered part of the "bootstrap" sequence.
    """

    #
    # 1st pass only matters for 'bootstrap' options : configuration file and logging
    #
    # Note : we don't let the parse_args method merge the Namespace into config yet,
    # because it would not be possible to make the difference between the default values
    # and the ones explictely given by the user
    # This is usefull for instance to throw an exception if a file given by the user doesn't exist, which is different than the default filename
    # 'config' is therefore the defaults overriden by user options while 'ns' has only user options
    ns = parser.parse_args(args=args)

    # Logging configuration
    configure_logging(ns.verbosity,debug=ns.debug)
    logging.debug( "Configuration for bootstrap : %s", repr(vars(ns)) )

    # Fills the config with user-defined default options from a config file
    try:
        # Allows config_file to be relative to the config_dir
        config.config_file = filter_files(
            [ns.config_file,
            os.path.join(ns.config_dir,"config.yml")],
            should_exist=True,
            fallback_to=1 )[0]
        logging.debug("Using config file %s",config.config_file)
        with open(config.config_file,'r') as file:
            # The FullLoader parameter handles the conversion from YAML
            # scalar values to Python the dictionary format
            try:
                # This is the required syntax in newer pyyaml distributions
                dictConfig = yaml.load(file, Loader=yaml.FullLoader)
            except AttributeError:
                # Some systems (e.g. raspbian) ship with an older version of pyyaml
                dictConfig = yaml.load(file)
            logging.debug("Successfully loaded configuration from %s : %s" % (config.config_file,repr(dictConfig)))
            config.__dict__.update(dictConfig)
    except OSError as e:
        # If it was a user-set option, stop here
        if ns.config_file == config.config_file:
            raise e
        else:
            logging.debug("Could not open %s ; no config file will be used",config.config_file)
            logging.debug(e, exc_info=True)
            pass
    # From here the config object has only the default values for all configuration options

    #
    # 2nd pass parses all options
    #
    # Updates again the existing config object with all parsed options
    config = parser.parse_args(args=args,namespace=config)
    # From the bootstrap parameters, only logging level may need to be read again
    configure_logging(config.verbosity,debug=config.debug)
    logging.debug( "Final configuration : %s", repr(vars(config)) )

    return config



# -*- coding: utf-8 -*-

import logging
import sys

from .chatter import Chatter


class ConsoleChatter(Chatter):
    """
        Bot engine that reads messages from a stream (stdin by default)
        and outputs to another stream (stdout by default)
    """

    def __init__( self, input=sys.stdin, output=sys.stdout ):
        self.input = input
        self.output = output
        self.exit = False

    def start( self, bot ):
        # TODO Do it asynchronous (rather than testing self.exit between each instruction)
        if self.exit:
            return
        for line in self.input:
            if self.exit:
                return
            logging.debug( "<<< %s", line )
            bot.onMessage( line.rstrip() )
            if self.exit:
                return

    def send( self, message ):
        print(message, file = self.output, flush=True)

    def stop( self ):
        self.exit = True
